#include "opencv2/opencv.hpp"

void printDeltaMilliSec(double tick_begin, double tick_end)
{
    double delta = 1000.0 * (tick_end - tick_begin) / cv::getTickFrequency();
    printf("%f [ms]\n", delta);
}

void onMouse(int event, int x, int y, int flags, void *param)
{
    if (event == CV_EVENT_LBUTTONDOWN) {
        cv::Point *tcenterp = (cv::Point *)param;
        tcenterp->x = x;
        tcenterp->y = y;
    }
}

int SSD(const cv::Mat& Tblock, const cv::Mat& Iblock)
{
    int ssd = 0;
    for (int j = 0; j < Tblock.rows; j++) {
        for (int i = 0; i < Tblock.cols; i++) {
            int d = Tblock.at<uchar>(j, i) - Iblock.at<uchar>(j, i);
            ssd += d * d;
        }
    }
    return ssd;
}

void trackBM(const cv::Mat& Icurr, const cv::Mat& Tblock, cv::Point& tcenter,
             cv::Size search_size)
{
    double tick_begin = (double)cv::getTickCount();

    int sxmin = tcenter.x - search_size.width / 2;
    int sxmax = tcenter.x + search_size.width / 2;
    int symin = tcenter.y - search_size.height / 2;
    int symax = tcenter.y + search_size.height / 2;

    if (sxmin < 0) { 
        sxmin = 0; 
    }
    if (sxmax >= Icurr.cols - Tblock.cols) { 
        sxmax = Icurr.cols - Tblock.cols - 1; 
    }
    if (symin < 0) { 
        symin = 0; 
    }
    if (symax >= Icurr.rows - Tblock.rows) { 
        symax = Icurr.rows - Tblock.rows - 1; 
    }

    int minssd = INT_MAX;
    cv::Point minloc = cv::Point(0, 0);
    for (int j = symin; j <= symax; j++) {
        for (int i = sxmin; i <= sxmax; i++) {
            cv::Mat Iblock = Icurr(cv::Rect(i, j, Tblock.cols, Tblock.rows));
            int ssd = SSD(Tblock, Iblock);
            if (ssd < minssd) {
                minssd = ssd;
                minloc.x = i;
                minloc.y = j;
            }
        }
    }
    tcenter.x = minloc.x + Tblock.cols / 2;
    tcenter.y = minloc.y + Tblock.rows / 2;

    double tick_end = (double)cv::getTickCount();

    printDeltaMilliSec(tick_begin, tick_end);
}

int main() {
    cv::namedWindow("output", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input, output;

    const int bwidth = 64;
    const int bheight = 64;

    cv::Point tcenter; // center of tracked region
    cap >> input;
    tcenter.x = input.cols / 2;
    tcenter.y = input.rows / 2;

    cv::setMouseCallback("output", onMouse, (void *)&tcenter);

    int search_margin = bwidth / 2;
    cv::createTrackbar("search margin", "output", &search_margin, bwidth);

    cv::Mat Iprev, Icurrent, Tblock;

    std::cout << "Click to position the template" << std::endl;

    while (1) {
        cap >> input;
        cv::cvtColor(input, Icurrent, CV_BGR2GRAY);
        if (Iprev.empty()) { // in case of the first frame
            Icurrent.copyTo(Iprev);
        }

        Tblock = Iprev(cv::Rect(tcenter.x - bwidth / 2,
                                tcenter.y - bheight / 2,
                            bwidth, bheight));
        trackBM(Icurrent, Tblock, tcenter, 
                cv::Size(bwidth + 2 * search_margin,
                         bheight + 2 * search_margin));
        Icurrent.copyTo(Iprev);

        input.copyTo(output);
        cv::rectangle(output,
                      cv::Point((int)tcenter.x - bwidth / 2,
                                (int)tcenter.y - bheight / 2),
                      cv::Point((int)tcenter.x + bwidth / 2,
                                (int)tcenter.y + bheight / 2), 
                      CV_RGB(255, 0, 0), 3);
        cv::rectangle(output,
                      cv::Point((int)tcenter.x - bwidth / 2 - search_margin,
                                (int)tcenter.y - bheight / 2 - search_margin),
                      cv::Point((int)tcenter.x + bwidth / 2 + search_margin,
                                (int)tcenter.y + bheight / 2 + search_margin), 
                      CV_RGB(0, 255, 0), 1);
        cv::imshow("output", output);
        if (cv::waitKey(30) > 0) {
            break;
        }
    }

    return 0;
}
