#include "opencv2/opencv.hpp"

void printDeltaMilliSec(double tick_begin, double tick_end)
{
    double delta = 1000.0 * (tick_end - tick_begin) / cv::getTickFrequency();
    printf("%f [ms]\n", delta);
}

void onMouse(int event, int x, int y, int flags, void *param)
{
    if (event == CV_EVENT_LBUTTONDOWN) {
        cv::Point2f *tcenterp = (cv::Point2f *)param;
        tcenterp->x = (float)x;
        tcenterp->y = (float)y;
    }
}

void pushImage(std::vector<cv::Mat>& matvec, const cv::Mat& image,
                   int scale, int offset, const std::string& text)
{
    cv::Mat image_r = image(cv::Rect(1, 1, image.cols - 2, image.rows - 2));
    // outmost pixels are ignored
    
    cv::Mat name = cv::Mat::zeros(20, image_r.cols, CV_16S);
    name -= 32768;
    cv::putText(name, text, cv::Point(1, 16),
                CV_FONT_HERSHEY_PLAIN, 1.0, cvScalar(32767));
    cv::Mat named_image;
    std::vector<cv::Mat> v;
    v.push_back(image_r * scale + offset);
    v.push_back(name);
    cv::vconcat(&v[0], 2, named_image);

    cv::Mat sep = cv::Mat::zeros(named_image.rows, 4, CV_16S);
    sep -= 32768;
    matvec.push_back(named_image);
    matvec.push_back(sep);
}

void trackLK(const cv::Mat& Icurr, const cv::Mat& Tblock, cv::Point2f& tcenter,
             cv::Mat& illust)
{
    double tick_begin = (double)cv::getTickCount();

    const int max_iter = 20;

    int h = Tblock.rows;
    int w = Tblock.cols;
    cv::Mat Tx(h, w, CV_16S);
    cv::Mat Ty(h, w, CV_16S);
    cv::Mat TxTx(h, w, CV_16S);
    cv::Mat TyTy(h, w, CV_16S);
    cv::Mat TxTy(h, w, CV_16S);

    float h11 = 0.0f, h12 = 0.0f, h22 = 0.0f;
    for (int j = 1; j < h - 1; j++) {
        for (int i = 1; i < w - 1; i++) {
            // outmost pixels are ignored

            Tx.at<short>(j, i) =
                (Tblock.at<short>(j, i + 1) - Tblock.at<short>(j, i - 1)) / 2;
            Ty.at<short>(j, i) =
                (Tblock.at<short>(j + 1, i) - Tblock.at<short>(j - 1, i)) / 2;
            TxTx.at<short>(j, i) = Tx.at<short>(j, i) * Tx.at<short>(j, i);
            TxTy.at<short>(j, i) = Tx.at<short>(j, i) * Ty.at<short>(j, i);
            TyTy.at<short>(j, i) = Ty.at<short>(j, i) * Ty.at<short>(j, i);

            h11 += TxTx.at<short>(j, i);
            h12 += TxTy.at<short>(j, i);
            h22 += TyTy.at<short>(j, i);
        }
    }
    cv::Mat H = (cv::Mat_<float>(2, 2) << h11, h12, h12, h22);

    bool is_converged = false;
    cv::Mat Iblock;
    cv::Mat TminI(h, w, CV_16S);
    cv::Mat TminI_Tx(h, w, CV_16S);
    cv::Mat TminI_Ty(h, w, CV_16S);

    int iter;
    for (iter = 0; iter < max_iter && !is_converged; iter++) {
        cv::getRectSubPix(Icurr, Tblock.size(), tcenter, Iblock);
        Iblock.convertTo(Iblock, CV_16S);

        float g1 = 0.0f, g2 = 0.0;
        for (int j = 1; j < h - 1; j++) {
            for (int i = 1; i < w - 1; i++) {
                TminI.at<short>(j, i) =
                    Tblock.at<short>(j, i) - Iblock.at<short>(j, i);
                TminI_Tx.at<short>(j, i) =
                    TminI.at<short>(j, i) * Tx.at<short>(j, i);
                TminI_Ty.at<short>(j, i) =
                    TminI.at<short>(j, i) * Ty.at<short>(j, i);

                g1 += TminI_Tx.at<short>(j, i);
                g2 += TminI_Ty.at<short>(j, i);
            }
        }

        cv::Mat g = (cv::Mat_<float>(2, 1) << g1, g2);
        cv::Mat sol;
        cv::solve(H, -g, sol, cv::DECOMP_SVD);
        
        float dx = sol.at<float>(0);
        float dy = sol.at<float>(1);
        tcenter.x -= dx;
        tcenter.y -= dy;

        is_converged = (dx * dx + dy * dy < 0.05) ? true : false;
    }

    double tick_end = (double)cv::getTickCount();
    printf("iter = %d\n", iter);
    printDeltaMilliSec(tick_begin, tick_end);


    std::vector<cv::Mat> matvec;
    pushImage(matvec, Tblock, 256, -32768, "T");
    pushImage(matvec, Iblock, 256, -32768, "I");
    pushImage(matvec, TminI, 256, 0, "T-I");
    pushImage(matvec, Tx, 256, 0, "Tx");
    pushImage(matvec, Ty, 256, 0, "Ty");
    pushImage(matvec, TxTx, 16, 0, "TxTx");
    pushImage(matvec, TxTy, 16, 0, "TxTy");
    pushImage(matvec, TyTy, 16, 0, "TyTy");
    pushImage(matvec, TminI_Tx, 16, 0, "(T-I)Tx");
    pushImage(matvec, TminI_Ty, 16, 0, "(T-I)Ty");
    cv::hconcat(&matvec[0], matvec.size(), illust);
}

int main() {
    cv::namedWindow("output", cv::WINDOW_AUTOSIZE);
    cv::namedWindow("illustration", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input, output, illust;

    const int bwidth = 64;
    const int bheight = 64;

    cv::Point2f tcenter; // center of tracked region
    cap >> input;
    tcenter.x = (float)input.cols / 2;
    tcenter.y = (float)input.rows / 2;

    cv::setMouseCallback("output", onMouse, (void *)&tcenter);

    cv::Mat Iprev, Icurrent, Tblock;

    while (1) {
        cap >> input;
        cv::cvtColor(input, Icurrent, CV_BGR2GRAY);
        if (Iprev.empty()) { // in case of the first frame
            Icurrent.copyTo(Iprev);
        }

        cv::getRectSubPix(Iprev, cv::Size(bwidth, bheight), tcenter, Tblock);
        Tblock.convertTo(Tblock, CV_16S);
        trackLK(Icurrent, Tblock, tcenter, illust);
        Icurrent.copyTo(Iprev);

        input.copyTo(output);
        cv::rectangle(output,
                      cv::Point((int)tcenter.x - bwidth / 2,
                                (int)tcenter.y - bheight / 2),
                      cv::Point((int)tcenter.x + bwidth / 2,
                                (int)tcenter.y + bheight / 2), 
                      CV_RGB(255, 0, 0), 3);
        cv::imshow("output", output);
        cv::imshow("illustration", illust);
        if (cv::waitKey(30) > 0) {
            break;
        }
    }

    return 0;
}
